#include <fstream.h>
#include <windows.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <math.h>

#define TRIANGLES    40
#define FOG_DENSITY  1.0
#define Z_RANGE		100
#define TEX_NAME	   "TEXTURE.BMP"
#define H_NAME	   	"MAP.BMP"
#define TEX_WIDTH		512
#define TEX_HEIGHT	512

UINT timer;
char *texture;
unsigned char *h_map;

void setDCPixelFormat(HDC hdc)
{
  static PIXELFORMATDESCRIPTOR pfd =
  {
    sizeof(PIXELFORMATDESCRIPTOR),          // sizeof this structure
    1,                                      // version number
    PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL
    | PFD_DOUBLEBUFFER,                     // flags
    PFD_TYPE_RGBA,                          // RGBA pixel values
    16,                                     // 24-bit color
    0, 0, 0, 0, 0, 0,                       // don't care about these
    0, 0,                                   // no alpha buffer
    0, 0, 0, 0, 0,                          // no accumulation buffer
    16,                                     // 32-bit depth buffer
    0,                                      // no stencil buffer
    0,                                      // no auxiliary buffers
    PFD_MAIN_PLANE,                         // layer type
    0,                                      // reserved(must be 0)
    0, 0, 0                                 // no layer masks
  };
  int pixelFormat;

  pixelFormat = ChoosePixelFormat(hdc, &pfd);
  SetPixelFormat(hdc, pixelFormat, &pfd);
  DescribePixelFormat(hdc, pixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
}

void InitializeGL()
{
  GLfloat fogColor[4] = {0.7, 0.7, 1.0, 1.0};

  glFrontFace(GL_CW);
  glCullFace(GL_BACK);
  glEnable(GL_CULL_FACE);
  glEnable(GL_SMOOTH);

  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);
  glClearColor(fogColor[0], fogColor[1], fogColor[2], fogColor[3]);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, TEX_WIDTH, TEX_HEIGHT, 0, GL_RGB, GL_UNSIGNED_BYTE, texture);
  glEnable(GL_TEXTURE_2D);
  glFogf(GL_FOG_MODE, GL_EXP2);
  glFogf(GL_FOG_DENSITY, FOG_DENSITY);
  glFogf(GL_FOG_END, Z_RANGE);
  glFogfv(GL_FOG_COLOR, fogColor);
  glEnable(GL_FOG);
}

void DrawSurface(GLfloat offset)
{
  GLfloat x, y = 0.0, C = 1.0 / (GLfloat)TRIANGLES;
  GLfloat last_x, last_y = y;
  long i1, i2, i3, i4;
  GLfloat h1, h2, h3, h4;

  glBegin(GL_TRIANGLES);
    for (int i = 0; i < TRIANGLES; i++)
    {
      x = 0.0;
      last_x = x;
      i2 = y * 512;
      i4 = last_y * 512;
      for (int j = 0; j < TRIANGLES * 2; j++)
      {
        i1 = long((x + offset) * 512) & 511;
        i3 = long((last_x + offset) * 512) & 511;

        h1 = GLfloat(h_map[i2 * 512 + i3]) / 1024;
        h2 = GLfloat(h_map[i2 * 512 + i1]) / 1024;
        h3 = GLfloat(h_map[i4 * 512 + i3]) / 1024;
        h4 = GLfloat(h_map[i4 * 512 + i1]) / 1024;

        glTexCoord2f(last_x + offset, y);
        glVertex3f(last_x - 1.0, y - 0.5, h1);
        glTexCoord2f(x + offset, y);
        glVertex3f(x - 1.0, y - 0.5, h2);
        glTexCoord2f(last_x + offset, last_y);
        glVertex3f(last_x - 1.0, last_y - 0.5, h3);

        glVertex3f(last_x - 1.0, last_y - 0.5, h3);
        glTexCoord2f(x + offset, y);
        glVertex3f(x - 1.0, y - 0.5, h2);
        glTexCoord2f(x + offset, last_y);
        glVertex3f(x - 1.0, last_y - 0.5, h4);

        last_x = x;
        x += C;
      }
      last_y = y;
      y += C;
    }
  glEnd();
}

void drawScene(HDC hDC)
{
  static GLfloat angle = -1.0;

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(0.0, -0.20 + sin(angle * 10) * 0.05, -1.0);
  glRotatef(-90.0, 1.0, 0.0, 0.0);
  glRotatef(90.0, 0.0, 0.0, 1.0);

  DrawSurface(angle);

  SwapBuffers(hDC);
  glFlush();

  angle += 0.025;
}

LONG WINAPI wndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  static HDC hdc;
  static HGLRC hrc;
  PAINTSTRUCT ps;
  GLdouble aspect;
  GLsizei width, height;

  switch(msg)
  {
    case WM_CREATE:
      hdc = GetDC(hWnd);
      setDCPixelFormat(hdc);
      hrc = wglCreateContext(hdc);
      wglMakeCurrent(hdc, hrc);
      InitializeGL();
      timer = SetTimer(hWnd, 1, 50, NULL);
      return 0;

    case WM_SIZE:
      width = (GLsizei)LOWORD(lParam);
      height = (GLsizei)HIWORD(lParam);
      if (!width || !height) return 0;

      aspect = (GLdouble)width /(GLdouble)height;
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(30.0, aspect, 0.1, 5.0);
      glViewport(0, 0, width, height);

      return 0;

    case WM_PAINT:
      BeginPaint(hWnd, &ps);
      drawScene(hdc);
      EndPaint(hWnd, &ps);
      return 0;

    case WM_DESTROY:
      wglMakeCurrent(NULL, NULL);
      wglDeleteContext(hrc);
      KillTimer(hWnd, timer);
      PostQuitMessage(0);
      return 0;

    case WM_TIMER:
      InvalidateRect(hWnd, NULL, FALSE);
      return 0;
  }

  return DefWindowProc(hWnd, msg, wParam, lParam);
}

int LoadFile(char *filename, long offs, long bytes, char *&buf)
{
  char err[100] = " ";
  ifstream in(filename, ios::binary);
  if (!in)
  {
    strcat(err, filename);
    strcat(err, "  ");
    MessageBox(0, err, "", MB_OK | MB_ICONSTOP);
    return 1;
  }
  in.seekg(offs, ios::beg);
  in.read(buf, bytes);
  in.close();

  return 0;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR cmdLine, int cmdShow)
{
  static char appName [] = "Sierra - by HackMaster";
  WNDCLASS wc;
  HWND hWnd;
  MSG msg;

  texture = new char[TEX_WIDTH * TEX_HEIGHT * 3];
  h_map = new char[262144];

  if (LoadFile(TEX_NAME, 0x36, TEX_WIDTH * TEX_HEIGHT * 3, texture)) return 1;
  if (LoadFile(H_NAME, 1078, 262144, h_map)) return 1;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc =(WNDPROC)wndProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground =(HBRUSH)(COLOR_WINDOW + 1);
  wc.lpszMenuName = (LPCTSTR)0;
  wc.lpszClassName = appName;

  RegisterClass(&wc);

  hWnd = CreateWindow(appName, appName,
                      WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
                      CW_USEDEFAULT, CW_USEDEFAULT, 320,
                      240, HWND_DESKTOP, NULL, hInstance, NULL);

  ShowWindow(hWnd, cmdShow);
  UpdateWindow(hWnd);

  while(GetMessage(&msg, NULL, 0, 0))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }

  return msg.wParam;
}

